import typing

import asyncpg  # type: ignore
from ponty import DoesNotExist

from pontypg.provider import lease_postgres_conn, defaultdb


async def fetchval(query: str, *a, dbname: str = defaultdb) -> typing.Any:
    async with lease_postgres_conn(dbname) as conn:
        return await conn.fetchval(query, *a)


async def fetchrow(query: str, *a, dbname: str = defaultdb) -> asyncpg.Record:
    async with lease_postgres_conn(dbname) as conn:
        record = await conn.fetchrow(query, *a)

    if not record:
        raise DoesNotExist

    return record


async def fetchrows(query: str, *a, dbname: str = defaultdb) -> list[asyncpg.Record]:
    async with lease_postgres_conn(dbname) as conn:
        return await conn.fetch(query, *a)




T = typing.TypeVar("T")


async def fetchone(query: str, *a, container: type[T], dbname: str = defaultdb) -> T:
    record = await fetchrow(query, *a, dbname=dbname)
    return container(**record)


async def fetchmany(query: str, *a, container: type[T], dbname: str = defaultdb) -> list[T]:
    records = await fetchrows(query, *a, dbname=dbname)
    return [container(**r) for r in records]



__all__ = [
    "postgres_provider", "lease_postgres_conn",
    "fetchval", "fetchrow", "fetchrows",
    "fetchone", "fetchmany",
    "error_trap",
]


from pontypg.errortrap import error_trap
from pontypg.fetch import fetchval, fetchrow, fetchrows, fetchone, fetchmany
from pontypg.provider import postgres_provider, lease_postgres_conn



from contextlib import asynccontextmanager
import typing

import asyncpg  # type: ignore
from ponty import Provider
from ponty.registry import Registry


_registry = Registry[asyncpg.pool.Pool]()
defaultdb: typing.Final[str] = "_default"


def postgres_provider(
    *,
    dbname: str = defaultdb,
    host: str,
    port: str = "5432",
    database: str,
    user: str,
    password: typing.Union[str, typing.Callable[[], str]] = "",
    min_size: int = 5,
    max_size: int = 20,
    command_timeout: typing.Optional[int] = 5,
) -> Provider:

    async def provider(_) -> typing.AsyncIterator[None]:
        pool = await asyncpg.create_pool(
            host=host,
            port=port,
            database=database,
            user=user,
            password=password,
            min_size=min_size,
            max_size=max_size,
            command_timeout=command_timeout,
        )

        _registry.add(dbname, pool)

        try:
            yield
        finally:
            await pool.close()

    return provider


@asynccontextmanager
async def lease_postgres_conn(dbname: str = defaultdb) -> typing.AsyncIterator[asyncpg.Connection]:
    pool = _registry.get(dbname)
    async with pool.acquire() as conn:
        yield conn



import functools
from typing import Awaitable, Callable, TypeVar
from typing_extensions import ParamSpec

import asyncpg.exceptions  # type: ignore
from ponty import ValidationError


P = ParamSpec("P")
R = TypeVar("R")


def error_trap(f: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]:
    @functools.wraps(f)
    async def wrapper(*a: P.args, **kw: P.kwargs) -> R:
        try:
            return await f(*a, **kw)

        except (
            asyncpg.exceptions.CheckViolationError,
            asyncpg.exceptions.ForeignKeyViolationError,
            asyncpg.exceptions.NotNullViolationError,
            asyncpg.exceptions.UniqueViolationError,
        ) as e:
            raise ValidationError(text=e.detail)

    return wrapper
